# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import re
from contextlib import contextmanager

from .scenario_tests import (create_random_name as create_random_name_base, RecordingProcessor)
from .scenario_tests.utilities import is_text_payload


def create_random_name(prefix='clitest', length=24):
    return create_random_name_base(prefix=prefix, length=length)


def find_recording_dir(test_file):
    """ Find the directory containing the recording of given test file based on current profile. """
    return os.path.join(os.path.dirname(test_file), 'recordings')


@contextmanager
def force_progress_logging():
    from io import StringIO
    import logging
    from knack.log import get_logger
    from .reverse_dependency import get_commands_loggers

    cmd_logger = get_commands_loggers()

    # register a progress logger handler to get the content to verify
    test_io = StringIO()
    test_handler = logging.StreamHandler(test_io)
    cmd_logger.addHandler(test_handler)
    old_cmd_level = cmd_logger.level
    cmd_logger.setLevel(logging.INFO)

    # this tells progress logger we are under verbose, so should log
    az_logger = get_logger()
    old_az_level = az_logger.handlers[0].level
    az_logger.handlers[0].level = logging.INFO

    yield test_io

    # restore old logging level and unplug the test handler
    cmd_logger.removeHandler(test_handler)
    cmd_logger.setLevel(old_cmd_level)
    az_logger.handlers[0].level = old_az_level


def _byte_to_str(byte_or_str):
    return str(byte_or_str, 'utf-8') if isinstance(byte_or_str, bytes) else byte_or_str


class StorageAccountKeyReplacer(RecordingProcessor):
    """Replace the access token for service principal authentication in a response body."""

    KEY_REPLACEMENT = 'veryFakedStorageAccountKey=='

    def __init__(self):
        self._activated = False
        self._candidates = []

    def reset(self):
        self._activated = False
        self._candidates = []

    def process_request(self, request):  # pylint: disable=no-self-use
        import re
        try:
            pattern = r"/providers/Microsoft\.Storage/storageAccounts/[^/]+/listKeys$"
            if re.search(pattern, request.path, re.I):
                self._activated = True
        except AttributeError:
            pass
        for candidate in self._candidates:
            if request.body:
                body_string = _byte_to_str(request.body)
                request.body = body_string.replace(candidate, self.KEY_REPLACEMENT)
        return request

    def process_response(self, response):
        if self._activated:
            import json
            try:
                body = json.loads(response['body']['string'])
                keys = body['keys']
                for key_entry in keys:
                    self._candidates.append(key_entry['value'])
                self._activated = False
            except (KeyError, ValueError, TypeError):
                pass
        for candidate in self._candidates:
            if response['body']['string']:
                body = response['body']['string']
                response['body']['string'] = _byte_to_str(body)
                response['body']['string'] = response['body']['string'].replace(candidate, self.KEY_REPLACEMENT)
        return response


class GraphClientPasswordReplacer(RecordingProcessor):
    """Replace the access token for service principal authentication in a response body."""

    PWD_REPLACEMENT = 'ReplacedSPPassword123*'

    def __init__(self):
        self._activated = False

    def reset(self):
        self._activated = False

    def process_request(self, request):  # pylint: disable=no-self-use
        import re
        import json

        try:
            # issue with how vcr.Request.body adds b' to text types if self.body is used.
            if request.body and self.PWD_REPLACEMENT in str(request.body):
                return request

            pattern = r"[^/]+/applications$"
            if re.search(pattern, request.path, re.I) and request.method.lower() == 'post':
                self._activated = True
                body = _byte_to_str(request.body)
                body = json.loads(body)
                for password_cred in body['passwordCredentials']:
                    if password_cred['value']:
                        body_string = _byte_to_str(request.body)
                        request.body = body_string.replace(password_cred['value'], self.PWD_REPLACEMENT)

        except (AttributeError, KeyError):
            pass

        return request

    def process_response(self, response):
        if self._activated:
            try:
                import json

                body = json.loads(response['body']['string'])
                for password_cred in body['passwordCredentials']:
                    password_cred['value'] = password_cred['value'] or self.PWD_REPLACEMENT

                response['body']['string'] = json.dumps(body)
                self._activated = False

            except (AttributeError, KeyError):
                pass

        return response


class MSGraphClientPasswordReplacer(RecordingProcessor):
    """Replace 'secretText' property in 'addPassword' API's response."""

    PWD_REPLACEMENT = 'replaced-microsoft-graph-password'

    def __init__(self):
        self._activated = False

    def reset(self):
        self._activated = False

    def process_request(self, request):
        if request.path.endswith('/addPassword') and request.method.lower() == 'post':
            self._activated = True
        return request

    def process_response(self, response):
        if self._activated:
            import json

            body = json.loads(response['body']['string'])
            body['secretText'] = self.PWD_REPLACEMENT

            response['body']['string'] = json.dumps(body)
            self._activated = False

        return response


class MSGraphNameReplacer(RecordingProcessor):
    """If a name appears in URL, it should be percent-encoded, e.g.
      - test-name+/ is encoded as test-name%2B%2F
      - example@example.com is encoded as example%40example.com

    Theoretically, GeneralNameReplacer should also follow this pattern, but since we haven't seen any ARM name that
    is percent-encoded, we only replace percent-encoded names for MS Graph for better test performance.
    """
    def __init__(self, test_name, mock_name):
        from urllib.parse import quote
        self.test_name = test_name
        self.test_name_encoded = quote(test_name, safe='')
        self.mock_name = mock_name
        self.mock_name_encoded = quote(mock_name, safe='')

    def process_request(self, request):
        request.uri = request.uri.replace(self.test_name_encoded, self.mock_name_encoded)

        if request.body:
            body = _byte_to_str(request.body)
            request.body = body.replace(self.test_name, self.mock_name)

        return request

    def process_response(self, response):
        if response['body']['string']:
            response['body']['string'] = response['body']['string'].replace(self.test_name, self.mock_name)
        return response


class EmailAddressReplacer(RecordingProcessor):
    """Replace email address like xxx@microsoft.com with test@example.com"""

    EMAIL_REPLACEMENT = 'test@example.com'

    def _replace_email_address(self, text):
        pattern = r'[\w.%#+-]+[%40|@|_]microsoft.com'
        return re.sub(pattern, self.EMAIL_REPLACEMENT, text)

    def process_request(self, request):
        request.uri = self._replace_email_address(request.uri)
        if request.body:
            try:
                body = _byte_to_str(request.body)
                request.body = self._replace_email_address(body)
            except UnicodeDecodeError:
                # If the body is not a string, we cannot decode it, so we skip the replacement
                pass
        return request

    def process_response(self, response):
        if response['body']['string']:
            try:
                body = _byte_to_str(response['body']['string'])
                response['body']['string'] = self._replace_email_address(body)
            except UnicodeDecodeError:
                # If the body is not a string, we cannot decode it, so we skip the replacement
                pass
        return response


class AADAuthRequestFilter(RecordingProcessor):
    """Remove oauth authentication requests and responses from recording.
    This is derived from OAuthRequestResponsesFilter.
    """
    def process_request(self, request):
        # filter AAD requests like:
        # Tenant discovery: GET
        # https://login.microsoftonline.com/54826b22-38d6-4fb2-bad9-b7b93a3e9c5a/v2.0/.well-known/openid-configuration
        # Get access token: POST
        # https://login.microsoftonline.com/54826b22-38d6-4fb2-bad9-b7b93a3e9c5a/oauth2/v2.0/token
        if request.uri.startswith('https://login.microsoftonline.com'):
            return None
        return request



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import re

from azure.cli.command_modules.acs._consts import (
    CONST_DEFAULT_NODE_OS_TYPE
)
from azure.cli.command_modules.acs.azurecontainerstorage._consts import (
    CONST_ACSTOR_ALL,
    CONST_ACSTOR_IO_ENGINE_LABEL_KEY,
    CONST_ACSTOR_IO_ENGINE_LABEL_VAL,
    CONST_ACSTOR_VERSION_V1,
    CONST_STORAGE_POOL_OPTION_NVME,
    CONST_STORAGE_POOL_OPTION_SSD,
    CONST_STORAGE_POOL_SKU_PREMIUM_LRS,
    CONST_STORAGE_POOL_SKU_PREMIUM_ZRS,
    CONST_STORAGE_POOL_SKU_PREMIUMV2_LRS,
    CONST_STORAGE_POOL_TYPE_AZURE_DISK,
    CONST_STORAGE_POOL_TYPE_ELASTIC_SAN,
    CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK,
)
from azure.cli.command_modules.acs.azurecontainerstorage._helpers import (
    get_vm_sku_details
)
from azure.cli.core.azclierror import (
    ArgumentUsageError,
    InvalidArgumentValueError,
    MutuallyExclusiveArgumentError,
    RequiredArgumentMissingError,
    UnknownError,
)
from knack.log import get_logger


elastic_san_supported_skus = [
    CONST_STORAGE_POOL_SKU_PREMIUM_LRS,
    CONST_STORAGE_POOL_SKU_PREMIUM_ZRS,
]

logger = get_logger(__name__)


def validate_disable_azure_container_storage_params_v1(  # pylint: disable=too-many-branches
    storage_pool_type,
    storage_pool_name,
    storage_pool_sku,
    storage_pool_option,
    storage_pool_size,
    nodepool_list,
    is_extension_installed,
    is_azureDisk_enabled,
    is_elasticSan_enabled,
    is_ephemeralDisk_localssd_enabled,
    is_ephemeralDisk_nvme_enabled,
    ephemeral_disk_volume_type,
    ephemeral_disk_nvme_perf_tier,
):
    if not is_extension_installed:
        raise InvalidArgumentValueError(
            'Invalid usage of --disable-azure-container-storage. '
            'Azure Container Storage is not enabled in the cluster.'
        )

    if not isinstance(storage_pool_type, str):
        raise InvalidArgumentValueError(
            f'Azure Container Storage has version {CONST_ACSTOR_VERSION_V1} installed. '
            'A pool type value must be specified when using --disable-azure-container-storage. '
            " Allowed values are: all, azureDisk, elasticSan, ephemeralDisk"
        )

    if storage_pool_name is not None:
        raise MutuallyExclusiveArgumentError(
            'Conflicting flags. Cannot define --storage-pool-name value '
            'when --disable-azure-container-storage is set.'
        )

    if storage_pool_sku is not None:
        raise MutuallyExclusiveArgumentError(
            'Conflicting flags. Cannot define --storage-pool-sku value '
            'when --disable-azure-container-storage is set.'
        )

    if storage_pool_size is not None:
        raise MutuallyExclusiveArgumentError(
            'Conflicting flags. Cannot define --storage-pool-size value '
            'when --disable-azure-container-storage is set.'
        )

    if ephemeral_disk_volume_type is not None:
        raise MutuallyExclusiveArgumentError(
            'Conflicting flags. Cannot define --ephemeral-disk-volume-type value '
            'when --disable-azure-container-storage is set.'
        )

    if ephemeral_disk_nvme_perf_tier is not None:
        raise MutuallyExclusiveArgumentError(
            'Conflicting flags. Cannot define --ephemeral-disk-nvme-perf-tier value '
            'when --disable-azure-container-storage is set.'
        )

    if storage_pool_option is not None:
        if storage_pool_type != CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
            raise ArgumentUsageError(
                'Cannot define --storage-pool-option value when '
                '--disable-azure-container-storage is not set '
                f'to {CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK}.'
            )

        if ((storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME and
           not is_ephemeralDisk_nvme_enabled) or
           (storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD and
           not is_ephemeralDisk_localssd_enabled)):
            raise InvalidArgumentValueError(
                'Invalid --storage-pool-option value since ephemeralDisk '
                f'of type {storage_pool_option} is not enabled in the cluster.'
            )
    else:
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and \
           is_ephemeralDisk_localssd_enabled and is_ephemeralDisk_nvme_enabled:
            raise RequiredArgumentMissingError(
                'Value of --storage-pool-option must be defined since ephemeralDisk of both '
                f'the types: {CONST_STORAGE_POOL_OPTION_NVME} and {CONST_STORAGE_POOL_OPTION_SSD} '
                'are enabled in the cluster.'
            )

    if nodepool_list is not None:
        raise MutuallyExclusiveArgumentError(
            'Conflicting flags. Cannot define --azure-container-storage-nodepools value '
            'when --disable-azure-container-storage is set.'
        )

    if storage_pool_type != CONST_ACSTOR_ALL:
        is_ephemeralDisk_enabled = is_ephemeralDisk_localssd_enabled or is_ephemeralDisk_nvme_enabled
        is_storagepool_type_not_active = (
            (storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK and
                not is_azureDisk_enabled) or
            (storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN and
                not is_elasticSan_enabled) or
            (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
                not is_ephemeralDisk_enabled)
        )

        if is_storagepool_type_not_active:
            raise ArgumentUsageError(
                'Invalid --disable-azure-container-storage value. '
                'Azure Container Storage is not enabled for storage pool '
                f'type {storage_pool_type} in the cluster.'
            )

        number_of_storagepool_types_active = 0
        if is_azureDisk_enabled:
            number_of_storagepool_types_active += 1
        if is_elasticSan_enabled:
            number_of_storagepool_types_active += 1
        if is_ephemeralDisk_nvme_enabled:
            number_of_storagepool_types_active += 1
        if is_ephemeralDisk_localssd_enabled:
            number_of_storagepool_types_active += 1

        number_of_storagepool_types_to_be_disabled = 0
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK or \
           storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN or \
           (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
                storage_pool_option != CONST_ACSTOR_ALL):
            number_of_storagepool_types_to_be_disabled = 1
        elif (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
                storage_pool_option == CONST_ACSTOR_ALL):
            if is_ephemeralDisk_nvme_enabled:
                number_of_storagepool_types_to_be_disabled += 1
            if is_ephemeralDisk_localssd_enabled:
                number_of_storagepool_types_to_be_disabled += 1

        if number_of_storagepool_types_active == number_of_storagepool_types_to_be_disabled:
            raise ArgumentUsageError(
                f'Since {storage_pool_type} is the only storage pool type enabled for Azure Container Storage, '
                'disabling the storage pool type will lead to disabling Azure Container Storage from the cluster. '
                f'To disable Azure Container Storage, set --disable-azure-container-storage to {CONST_ACSTOR_ALL}.'
            )


def validate_enable_azure_container_storage_v1_params(  # pylint: disable=too-many-locals,too-many-branches
    storage_pool_type,
    storage_pool_name,
    storage_pool_sku,
    storage_pool_option,
    storage_pool_size,
    nodepool_list,
    agentpool_details,
    is_v1_extension_installed,
    is_v2_extension_installed,
    v2_extension_version,
    is_azureDisk_enabled,
    is_elasticSan_enabled,
    is_ephemeralDisk_localssd_enabled,
    is_ephemeralDisk_nvme_enabled,
    ephemeral_disk_volume_type,
    ephemeral_disk_nvme_perf_tier,
    existing_ephemeral_disk_volume_type,
    existing_ephemeral_disk_nvme_perf_tier,
):

    if is_v2_extension_installed:
        #  Hardcoding this error message to version 1 because this is only valid for v1
        raise InvalidArgumentValueError(
            'Failed to enable Azure Container Storage version 1 as Azure Container Storage version '
            f'{v2_extension_version} is already installed on the cluster. Try enabling this version on another '
            'cluster. You can also enable this version by first disabling the existing installation of '
            'Azure Container Storage by running --disable-azure-container-storage. '
            'Note that disabling can impact existing workloads that depend on Azure Container Storage.'
        )

    if storage_pool_type not in [
        CONST_STORAGE_POOL_TYPE_AZURE_DISK,
        CONST_STORAGE_POOL_TYPE_ELASTIC_SAN,
        CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK
    ]:
        raise InvalidArgumentValueError(
            f'Storage pool type value must be specified for --enable-azure-container-storage '
            'when enabling Azure Container Storage v1. '
            'Supported values are '
            f'{CONST_STORAGE_POOL_TYPE_AZURE_DISK}, '
            f'{CONST_STORAGE_POOL_TYPE_ELASTIC_SAN}, '
            f'{CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK}.'
        )

    if storage_pool_name is not None:
        pattern = r'[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*'
        is_pool_name_valid = re.fullmatch(pattern, storage_pool_name)
        if not is_pool_name_valid:
            raise InvalidArgumentValueError(
                "Invalid --storage-pool-name value. "
                "Accepted values are lowercase alphanumeric characters, "
                "'-' or '.', and must start and end with an alphanumeric character.")

    if storage_pool_sku is not None:
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
            raise ArgumentUsageError(
                'Cannot set --storage-pool-sku when --enable-azure-container-storage is ephemeralDisk.'
            )
        if (
            storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN and
            storage_pool_sku not in elastic_san_supported_skus
        ):
            supported_skus_str = ", ".join(elastic_san_supported_skus)
            raise ArgumentUsageError(
                'Invalid --storage-pool-sku value. '
                f'Supported value for --storage-pool-sku are {supported_skus_str} '
                'when --enable-azure-container-storage is set to elasticSan.'
            )

    if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
        if ephemeral_disk_volume_type is None and ephemeral_disk_nvme_perf_tier is None:
            if storage_pool_option is None:
                raise RequiredArgumentMissingError(
                    'Value of --storage-pool-option must be defined when '
                    '--enable-azure-container-storage is set to ephemeralDisk.'
                )
        else:
            required_type_installed_for_disk_vol_type = False
            required_type_installed_for_nvme_perf_tier = False

            if ephemeral_disk_volume_type is not None:
                required_type_installed_for_disk_vol_type = is_v1_extension_installed and \
                    (is_ephemeralDisk_localssd_enabled or is_ephemeralDisk_nvme_enabled)

            if ephemeral_disk_nvme_perf_tier is not None:
                required_type_installed_for_nvme_perf_tier = is_v1_extension_installed and \
                    is_ephemeralDisk_nvme_enabled

            if storage_pool_option is None:
                option_needed_by_dependent_params = []
                supported_option_needed = ''
                if ephemeral_disk_volume_type is not None and not required_type_installed_for_disk_vol_type:
                    option_needed_by_dependent_params.append('--ephemeral-disk-volume-type')
                    supported_option_needed = f'{CONST_STORAGE_POOL_OPTION_NVME} or {CONST_STORAGE_POOL_OPTION_SSD}'
                if ephemeral_disk_nvme_perf_tier is not None and not required_type_installed_for_nvme_perf_tier:
                    option_needed_by_dependent_params.append('--ephemeral-disk-nvme-perf-tier')
                    supported_option_needed = {CONST_STORAGE_POOL_OPTION_NVME}

                if len(option_needed_by_dependent_params) > 0:
                    params_requiring_options = ', '.join(option_needed_by_dependent_params)
                    raise ArgumentUsageError(
                        f'Cannot set {params_requiring_options} along with --enable-azure-container-storage '
                        f'when storage pool type: {CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK} option: '
                        f'{supported_option_needed} is not enabled for Azure Container Storage. '
                        'Enable the option using --storage-pool-option.'
                    )
            else:
                if ephemeral_disk_nvme_perf_tier is not None and not required_type_installed_for_nvme_perf_tier and \
                   storage_pool_option != CONST_STORAGE_POOL_OPTION_NVME:
                    raise ArgumentUsageError(
                        'Cannot set --ephemeral-disk-nvme-perf-tier along with --enable-azure-container-storage '
                        f'when storage pool type: {CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK} option: '
                        f'{CONST_STORAGE_POOL_OPTION_NVME} is not enabled for Azure Container Storage. '
                        'Enable the option using --storage-pool-option.'
                    )

            if (storage_pool_name is not None or storage_pool_sku is not None or
               storage_pool_size is not None or nodepool_list is not None):
                # pylint: disable=too-many-boolean-expressions
                if (ephemeral_disk_volume_type is not None and
                    required_type_installed_for_disk_vol_type and ephemeral_disk_nvme_perf_tier is None) or \
                    (ephemeral_disk_volume_type is None and
                     ephemeral_disk_nvme_perf_tier is not None and required_type_installed_for_nvme_perf_tier) or \
                    (ephemeral_disk_volume_type is not None and required_type_installed_for_disk_vol_type and
                     ephemeral_disk_nvme_perf_tier is not None and required_type_installed_for_nvme_perf_tier):
                    enabled_options_arr = []
                    if is_ephemeralDisk_nvme_enabled:
                        enabled_options_arr.append(CONST_STORAGE_POOL_OPTION_NVME)
                    if is_ephemeralDisk_localssd_enabled:
                        enabled_options_arr.append(CONST_STORAGE_POOL_OPTION_SSD)
                    enabled_options = ', '.join(enabled_options_arr)

                    params_defined_arr = []
                    if storage_pool_size is not None:
                        params_defined_arr.append('--storage-pool-size')
                    if storage_pool_sku is not None:
                        params_defined_arr.append('--storage-pool-sku')
                    if storage_pool_name is not None:
                        params_defined_arr.append('--storage-pool-name')
                    if nodepool_list is not None:
                        params_defined_arr.append('--azure-container-storage-nodepools')
                    params = ', '.join(params_defined_arr)

                    flags_defined_arr = []
                    if ephemeral_disk_volume_type is not None:
                        flags_defined_arr.append('--ephemeral-disk-volume-type')
                    if ephemeral_disk_nvme_perf_tier is not None:
                        flags_defined_arr.append('--ephemeral-disk-nvme-perf-tier')

                    flags_defined = ', '.join(flags_defined_arr)

                    raise ArgumentUsageError(
                        f'Cannot set {params} for creation of a new storage pool while setting value for '
                        f'{flags_defined} since {CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK} storage pool type is '
                        f'already enabled for storage pool option {enabled_options}.'
                    )
            else:
                if required_type_installed_for_disk_vol_type and \
                   ephemeral_disk_volume_type is not None and \
                   ephemeral_disk_nvme_perf_tier is None and \
                   existing_ephemeral_disk_volume_type.lower() == ephemeral_disk_volume_type.lower():
                    raise InvalidArgumentValueError(
                        'Azure Container Storage is already configured with --ephemeral-disk-volume-type '
                        f'value set to {existing_ephemeral_disk_volume_type}.'
                    )

                if required_type_installed_for_nvme_perf_tier and \
                   ephemeral_disk_nvme_perf_tier is not None and \
                   ephemeral_disk_volume_type is None and \
                   existing_ephemeral_disk_nvme_perf_tier.lower() == ephemeral_disk_nvme_perf_tier.lower():
                    raise InvalidArgumentValueError(
                        'Azure Container Storage is already configured with --ephemeral-disk-nvme-perf-tier '
                        f'value set to {existing_ephemeral_disk_nvme_perf_tier}.'
                    )

                # pylint: disable=too-many-boolean-expressions
                if required_type_installed_for_disk_vol_type and \
                   ephemeral_disk_volume_type is not None and \
                   existing_ephemeral_disk_volume_type.lower() == ephemeral_disk_volume_type.lower() and \
                   required_type_installed_for_nvme_perf_tier and \
                   ephemeral_disk_nvme_perf_tier is not None and \
                   existing_ephemeral_disk_nvme_perf_tier.lower() == ephemeral_disk_nvme_perf_tier.lower():
                    raise InvalidArgumentValueError(
                        'Azure Container Storage is already configured with --ephemeral-disk-volume-type '
                        f'value set to {existing_ephemeral_disk_volume_type} and --ephemeral-disk-nvme-perf-tier '
                        f'value set to {existing_ephemeral_disk_nvme_perf_tier}.'
                    )

        if storage_pool_option == CONST_ACSTOR_ALL:
            raise InvalidArgumentValueError(
                f'Cannot set --storage-pool-option value as {CONST_ACSTOR_ALL} '
                'when --enable-azure-container-storage is set.'
            )
    else:
        if storage_pool_option is not None:
            raise ArgumentUsageError(
                'Cannot set --storage-pool-option when --enable-azure-container-storage is not ephemeralDisk.'
            )

        if ephemeral_disk_volume_type is not None:
            raise ArgumentUsageError(
                'Cannot set --ephemeral-disk-volume-type when --enable-azure-container-storage is not ephemeralDisk.'
            )

        if ephemeral_disk_nvme_perf_tier is not None:
            raise ArgumentUsageError(
                'Cannot set --ephemeral-disk-nvme-perf-tier when --enable-azure-container-storage is not ephemeralDisk.'
            )

    _validate_storage_pool_size(storage_pool_size, storage_pool_type)

    _validate_nodepools(
        nodepool_list,
        agentpool_details,
        storage_pool_type,
        storage_pool_option,
        storage_pool_sku,
        is_v1_extension_installed
    )

    if is_v1_extension_installed:
        if (is_azureDisk_enabled and
           storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK) or \
           (is_elasticSan_enabled and
           storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN):
            raise ArgumentUsageError(
                'Invalid --enable-azure-container-storage value. '
                'Azure Container Storage is already enabled for storage pool type '
                f'{storage_pool_type} in the cluster.'
            )

        # pylint: disable=too-many-boolean-expressions
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and \
           ephemeral_disk_volume_type is None and \
           ((is_ephemeralDisk_nvme_enabled and
            storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME and
            ephemeral_disk_nvme_perf_tier is None) or
            (is_ephemeralDisk_localssd_enabled and
                storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD)):
            ephemeral_disk_type_installed = CONST_STORAGE_POOL_OPTION_SSD if \
                is_ephemeralDisk_localssd_enabled else CONST_STORAGE_POOL_OPTION_NVME
            raise ArgumentUsageError(
                'Invalid --enable-azure-container-storage value. '
                'Azure Container Storage is already enabled for storage pool type '
                f'{CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK} and option {ephemeral_disk_type_installed} '
                'in the cluster.'
            )


def validate_enable_azure_container_storage_params(
    enablement_option,
    is_extension_installed,
    is_ephemeral_disk_enabled,
    is_elastic_san_enabled,
    is_v1_extension_installed,
    v1_extension_version,
    storage_pool_name,
    storage_pool_sku,
    storage_pool_option,
    storage_pool_size
):
    if is_v1_extension_installed:
        raise InvalidArgumentValueError(
            f'Failed to enable the latest version of Azure Container Storage as version {v1_extension_version} '
            'is already installed on the cluster. Try enabling Azure Container Storage in another cluster. '
            'You can also enable the latest version by first disabling the existing installation using '
            '--disable-azure-container-storage all. Note that disabling can impact existing workloads '
            'that depend on Azure Container Storage.'
        )

    if not enablement_option or enablement_option is True:
        if is_extension_installed:
            raise InvalidArgumentValueError(
                'Cannot enable Azure Container Storage as it is already enabled on the cluster.'
            )
    else:
        enablement_option_arr = enablement_option if isinstance(enablement_option, list) else [enablement_option]
        enable_ephemeral_disk = CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK in enablement_option
        enable_elastic_san = CONST_STORAGE_POOL_TYPE_ELASTIC_SAN in enablement_option
        for option in enablement_option_arr:
            if option not in [CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK, CONST_STORAGE_POOL_TYPE_ELASTIC_SAN]:
                raise InvalidArgumentValueError(
                    f"Unsupported storage option '{option}'. "
                    f"Supported values are '{CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK}' "
                    f"and '{CONST_STORAGE_POOL_TYPE_ELASTIC_SAN}'."
                )
        if is_ephemeral_disk_enabled == enable_ephemeral_disk and is_elastic_san_enabled == enable_elastic_san:
            options_display = "', '".join(enablement_option_arr)
            raise InvalidArgumentValueError(
                f"Cannot enable the requested storage options ('{options_display}') "
                "as they are already enabled on the cluster."
            )

    if storage_pool_name is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-name value. '
            f'Please remove --storage-pool-name {storage_pool_name} from the command and try again.'
        )

    if storage_pool_sku is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-sku value. '
            f'Please remove --storage-pool-sku {storage_pool_sku} from the command and try again.'
        )

    if storage_pool_option is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-option value. '
            f'Please remove --storage-pool-option {storage_pool_option} from the command and try again.'
        )

    if storage_pool_size is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-size value. '
            f'Please remove --storage-pool-size {storage_pool_size} from the command and try again.'
        )


def validate_disable_azure_container_storage_params(
    disablement_option,
    is_extension_installed,
    is_ephemeral_disk_enabled,
    is_elastic_san_enabled,
    storage_pool_name,
    storage_pool_sku,
    storage_pool_option,
    storage_pool_size
):
    if not is_extension_installed:
        raise InvalidArgumentValueError(
            'Cannot disable Azure Container Storage as it could not be found on the cluster.'
        )

    if disablement_option and disablement_option not in [True, CONST_ACSTOR_ALL]:
        actionable = False
        disablement_option_arr = disablement_option if isinstance(disablement_option, list) else [disablement_option]
        for disable_option in disablement_option_arr:
            if disable_option == CONST_ACSTOR_ALL:
                actionable = True
            elif disable_option == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
                actionable = actionable or is_ephemeral_disk_enabled
            elif disable_option == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN:
                actionable = actionable or is_elastic_san_enabled
            else:
                raise InvalidArgumentValueError(
                    f"Cannot disable unsupported storage option '{disable_option}'. "
                    f"Supported values are '{CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK}', "
                    f"'{CONST_STORAGE_POOL_TYPE_ELASTIC_SAN}' and '{CONST_ACSTOR_ALL}'."
                )
        if not actionable:
            options_display = "', '".join(disablement_option_arr)
            raise InvalidArgumentValueError(
                f"Cannot disable the requested storage options ('{options_display}') "
                "as they could not be found on the cluster."
            )

    if storage_pool_name is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-name value. '
            f'Please remove --storage-pool-name {storage_pool_name} from the command and try again.'
        )

    if storage_pool_sku is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-sku value. '
            f'Please remove --storage-pool-sku {storage_pool_sku} from the command and try again.'
        )

    if storage_pool_option is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-option value. '
            f'Please remove --storage-pool-option {storage_pool_option} from the command and try again.'
        )

    if storage_pool_size is not None:
        raise InvalidArgumentValueError(
            'The latest version of Azure Container Storage does not '
            'require or support a --storage-pool-size value. '
            f'Please remove --storage-pool-size {storage_pool_size} from the command and try again.'
        )


# _Validate_storage_pool_size validates that the storage_pool_size is
# string of a combination of a float number immediately followed by
# Ti or Gi e.g. 2Ti, 512Gi, 1.5Ti. The function also validates that the
# minimum storage pool size for an elastic san storagepool should be 1Ti.
def _validate_storage_pool_size(storage_pool_size, storage_pool_type):
    if storage_pool_size is not None:
        pattern = r'^\d+(\.\d+)?[GT]i$'
        match = re.match(pattern, storage_pool_size)
        if match is None:
            raise ArgumentUsageError(
                'Value for --storage-pool-size should be defined '
                'with size followed by Gi or Ti e.g. 512Gi or 2Ti.'
            )
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN:
            pool_size_qty = float(storage_pool_size[:-2])
            pool_size_unit = storage_pool_size[-2:]

            if (
                (pool_size_unit == "Gi" and pool_size_qty < 1024) or
                (pool_size_unit == "Ti" and pool_size_qty < 1)
            ):
                raise ArgumentUsageError(
                    'Value for --storage-pool-size must be at least 1Ti when '
                    '--enable-azure-container-storage is elasticSan.'
                )

        elif storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
            logger.warning(
                'Storage pools using Ephemeral disk use all capacity available on the local device. '
                ' --storage-pool-size will be ignored.'
            )


def _validate_nodepools(  # pylint: disable=too-many-branches,too-many-locals
    nodepool_list,
    agentpool_details,
    storage_pool_type,
    storage_pool_option,
    storage_pool_sku,
    is_extension_installed,
):
    nodepool_arr = []
    insufficient_core_error = (
        'Cannot operate Azure Container Storage on a node pool consisting of '
        'nodes with cores less than 4. Node pool: {0} with node size: {1} '
        'which is assigned for Azure Container Storage has nodes with {2} cores.'
    )
    if is_extension_installed:
        if nodepool_list is not None:
            raise ArgumentUsageError(
                "Cannot set --azure-container-storage-nodepools while using "
                "--enable-azure-container-storage to enable a type of storage pool "
                "in a cluster where Azure Container Storage is already installed. "
                "Use 'az aks nodepool' to label the node pool instead."
            )

        if agentpool_details is not None:
            for nodepool_name, agentpool in agentpool_details.items():
                node_labels = agentpool.get("node_labels")
                if node_labels is not None and \
                   node_labels.get(CONST_ACSTOR_IO_ENGINE_LABEL_KEY) is not None:
                    nodepool_arr.append(nodepool_name)

        if len(nodepool_arr) == 0:
            raise ArgumentUsageError(
                f'Cannot enable Azure Container Storage storage pool of type {storage_pool_type} '
                'since none of the nodepools in the cluster are labelled for Azure Container Storage.'
            )

        insufficient_core_error = (
            f'Cannot enable Azure Container Storage storage pool type: {storage_pool_type} '
            'on a node pool consisting of nodes with cores less than 4. '
            'Node pool: {0} with node size: {1} has nodes with {2} cores. '
            f'Remove the label {CONST_ACSTOR_IO_ENGINE_LABEL_KEY}={CONST_ACSTOR_IO_ENGINE_LABEL_VAL} '
            'from the node pool and use node pools which has nodes with 4 or more cores and try again.'
        )
    else:
        if not nodepool_list:
            agentpool_names_str = ', '.join(agentpool_details.keys())
            raise RequiredArgumentMissingError(
                'Multiple node pools present. Please define the node pools on which you want '
                'to enable Azure Container Storage using --azure-container-storage-nodepools.'
                f'\nNode pools available in the cluster are: {agentpool_names_str}.'
                '\nAborting Azure Container Storage operation.'
            )
        pattern = r'^[a-z][a-z0-9]*(?:,[a-z][a-z0-9]*)*$'
        if re.fullmatch(pattern, nodepool_list) is None:
            raise InvalidArgumentValueError(
                'Invalid --azure-container-storage-nodepools value. '
                'Accepted value is a comma separated string of valid node pool '
                'names without any spaces.\nA valid node pool name may only contain lowercase '
                'alphanumeric characters and must begin with a lowercase letter.'
            )
        nodepool_arr = nodepool_list.split(',')

    nvme_nodepool_found = False
    available_node_count = 0
    multi_zoned_cluster = False

    for pool_name in nodepool_arr:
        agentpool = agentpool_details.get(pool_name)
        if agentpool is not None:
            os_type = agentpool.get("os_type")
            if os_type is not None and os_type.lower() != CONST_DEFAULT_NODE_OS_TYPE.lower():
                raise InvalidArgumentValueError(
                    f'Azure Container Storage can be enabled only on {CONST_DEFAULT_NODE_OS_TYPE} nodepools. '
                    f'Node pool: {pool_name}, os type: {os_type} does not meet the criteria.'
                )
            mode = agentpool.get("mode")
            node_taints = agentpool.get("node_taints")
            if mode is not None and mode.lower() == "system" and node_taints is not None:
                critical_taint = "CriticalAddonsOnly=true:NoSchedule"
                if critical_taint.casefold() in (taint.casefold() for taint in node_taints):
                    raise InvalidArgumentValueError(
                        f'Unable to install Azure Container Storage on system nodepool: {pool_name} '
                        f'since it has a taint {critical_taint}. Remove the taint from the node pool '
                        'and retry the Azure Container Storage operation.'
                    )
            vm_size = agentpool.get("vm_size")
            if vm_size is not None and vm_size != "":
                cpu_value, nvme_enabled = get_vm_sku_details(vm_size.lower())
                if cpu_value is None or nvme_enabled is None:
                    raise UnknownError(
                        f'Unable to find details for virtual machine size {vm_size}.'
                    )
                if cpu_value < 0:
                    raise UnknownError(
                        f'Unable to determine number of cores in node pool: {pool_name}, node size: {vm_size}'
                    )
                if cpu_value < 4:
                    raise InvalidArgumentValueError(insufficient_core_error.format(pool_name, vm_size, cpu_value))

                nvme_nodepool_found = nvme_nodepool_found or nvme_enabled

            node_count = agentpool.get("count")
            if node_count is not None:
                available_node_count = available_node_count + node_count

            zoned_nodepool = agentpool.get("zoned")
            if zoned_nodepool:
                multi_zoned_cluster = True
        else:
            agentpool_names_str = ", ".join(agentpool_details.keys())
            nodepool_str = "Node pool"
            if len(agentpool_details.keys()) > 1:
                nodepool_str = "Node pools"
            raise InvalidArgumentValueError(
                f'Node pool: {pool_name} not found. '
                'Please provide a comma separated string of existing node pool names '
                'in --azure-container-storage-nodepools.'
                f'\n{nodepool_str} available in the cluster: {agentpool_names_str}.'
                '\nAborting installation of Azure Container Storage.'
            )

    if available_node_count < 3:
        raise UnknownError(
            'Insufficient nodes present. Azure Container Storage requires atleast 3 nodes to be enabled.'
        )

    if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and \
       storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME and \
       not nvme_nodepool_found:
        raise ArgumentUsageError(
            f'Cannot set --storage-pool-option as {CONST_STORAGE_POOL_OPTION_NVME} '
            'as none of the node pools can support ephemeral NVMe disk.'
        )

    if storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK and \
       storage_pool_sku == CONST_STORAGE_POOL_SKU_PREMIUMV2_LRS and \
       not multi_zoned_cluster:
        raise ArgumentUsageError(
            f'Cannot set --storage-pool-sku as {CONST_STORAGE_POOL_SKU_PREMIUMV2_LRS} '
            'as none of the node pools are zoned. Please add a zoned node pool and '
            'try again.'
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from unittest import mock

from azure.cli.core.profiles import (ResourceType, PROFILE_TYPE, CustomResourceType,
                                     get_api_version, supported_api_version, register_resource_type)
from azure.cli.core.profiles._shared import APIVersionException, get_versioned_sdk_path
from azure.cli.core.cloud import Cloud
from azure.cli.core.mock import DummyCli


class TestAPIProfiles(unittest.TestCase):

    def test_get_api_version(self):
        # Can get correct resource type API version
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertEqual(get_api_version(cli, ResourceType.MGMT_STORAGE), '2020-10-10')

    def test_get_api_version_semver(self):
        # Can get correct resource type API version if semver used
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertEqual(get_api_version(cli, ResourceType.MGMT_KEYVAULT), '7.0')

    def test_get_api_version_invalid_rt(self):
        # Resource Type not in profile
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            with self.assertRaises(APIVersionException):
                get_api_version(cli, ResourceType.MGMT_COMPUTE)

    def test_get_api_version_invalid_active_profile(self):
        # The active profile is not in our profile dict
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='not-a-real-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            with self.assertRaises(APIVersionException):
                get_api_version(cli, ResourceType.MGMT_STORAGE)

    def test_supported_api_version_invalid_profile_name(self):
        # Invalid name for the profile name
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='not-a-real-profile')
        with self.assertRaises(ValueError):
            supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-01')

    def test_get_api_version_invalid_rt_2(self):
        # None is not a valid resource type
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            with self.assertRaises(APIVersionException):
                get_api_version(cli, None)

    def test_supported_api_profile_no_constraints(self):
        # At least a min or max version must be specified
        cli = DummyCli()
        with self.assertRaises(ValueError):
            supported_api_version(cli, PROFILE_TYPE)

    def test_supported_api_profile_min_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2000-01-01-profile')
        self.assertTrue(supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-01'))

    def test_supported_api_profile_min_constraint_not_supported(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2000-01-01-profile-preview')
        self.assertFalse(supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-02'))

    def test_supported_api_profile_min_max_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2000-01-01-profile')
        self.assertTrue(supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-01', max_api='2000-01-01'))

    def test_supported_api_profile_max_constraint_not_supported(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2000-01-01-profile')
        self.assertFalse(supported_api_version(cli, PROFILE_TYPE, max_api='1999-12-30'))

    def test_supported_api_profile_preview_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2000-01-01-profile')
        self.assertTrue(supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-01-preview'))

    def test_supported_api_profile_preview_constraint_in_profile(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2000-01-01-profile-preview')
        self.assertFalse(supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-01'))

    def test_supported_api_profile_latest(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='latest')
        self.assertTrue(supported_api_version(cli, PROFILE_TYPE, min_api='2000-01-01'))

    def test_supported_api_version_no_constraints(self):
        # At least a min or max version must be specified
        cli = DummyCli()
        with self.assertRaises(ValueError):
            supported_api_version(cli, ResourceType.MGMT_STORAGE)

    def test_supported_api_version_min_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_STORAGE, min_api='2000-01-01'))

    def test_supported_api_version_min_constraint_semver(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, min_api='6.0'))

    def test_supported_api_version_min_constraint_mixed_type(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, min_api='2016-06-04'))

    def test_supported_api_version_max_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_STORAGE, max_api='2021-01-01'))

    def test_supported_api_version_max_constraint_semver(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, max_api='8.0'))

    def test_supported_api_version_max_constraint_mixed_type(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '2016-06-04'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, max_api='8.0'))

    def test_supported_api_version_min_max_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(
                supported_api_version(cli, ResourceType.MGMT_STORAGE, min_api='2020-01-01', max_api='2021-01-01'))

    def test_supported_api_version_min_max_constraint_semver(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(
                supported_api_version(cli, ResourceType.MGMT_KEYVAULT, min_api='6.0', max_api='8.0'))

    def test_supported_api_version_max_constraint_not_supported(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertFalse(supported_api_version(cli, ResourceType.MGMT_STORAGE, max_api='2019-01-01'))

    def test_supported_api_version_max_constraint_not_supported_semver(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertFalse(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, max_api='6.0'))

    def test_supported_api_version_max_constraint_not_supported_mixed_type(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertFalse(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, max_api='2016-07-01'))

    def test_supported_api_version_min_constraint_not_supported(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertFalse(supported_api_version(cli, ResourceType.MGMT_STORAGE, min_api='2021-01-01'))

    def test_supported_api_version_min_constraint_not_supported_semver(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_KEYVAULT: '7.0'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertFalse(supported_api_version(cli, ResourceType.MGMT_KEYVAULT, min_api='8.0'))

    def test_supported_api_version_min_constraint_not_supported_mixed_type(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2016-06-04'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertFalse(supported_api_version(cli, ResourceType.MGMT_STORAGE, min_api='8.0'))

    def test_supported_api_version_preview_constraint(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10-preview'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertTrue(supported_api_version(cli, ResourceType.MGMT_STORAGE, min_api='2020-01-01'))

    def test_supported_api_version_invalid_rt_for_profile(self):
        cli = DummyCli()
        cli.cloud = Cloud('TestCloud', profile='2017-01-01-profile')
        test_profile = {'2017-01-01-profile': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            with self.assertRaises(APIVersionException):
                supported_api_version(cli, ResourceType.MGMT_COMPUTE, min_api='2020-01-01')

    def test_register_resource_type_valid(self):
        test_profile = {'latest': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles.API_PROFILES', test_profile):
            custom_rt = CustomResourceType('mysdkpath', 'myClient')
            register_resource_type('latest', custom_rt, '2020-12-12-preview')
            from azure.cli.core.profiles import API_PROFILES
            self.assertDictEqual(API_PROFILES, {'latest': {ResourceType.MGMT_STORAGE: '2020-10-10',
                                                           custom_rt: '2020-12-12-preview'}})

    def test_register_resource_type_bad_profile(self):
        test_profile = {'latest': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles.API_PROFILES', test_profile):
            custom_rt = CustomResourceType('mysdkpath', 'myClient')
            with self.assertRaises(ValueError):
                register_resource_type('doesnotexist', custom_rt, '2020-12-12-preview')

    def test_register_resource_type_bad_rt_string(self):
        test_profile = {'latest': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles.API_PROFILES', test_profile):
            with self.assertRaises(TypeError):
                register_resource_type('doesnotexist', 'ResourceType.MGMT_STORAGE', '2020-12-12-preview')

    def test_register_resource_type_bad_rt_attempted_override(self):
        test_profile = {'latest': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles.API_PROFILES', test_profile):
            with self.assertRaises(TypeError):
                register_resource_type('doesnotexist', ResourceType.MGMT_STORAGE, '2020-12-12-preview')

    def test_get_versioned_sdk_path_date(self):
        test_profile = {'latest': {ResourceType.MGMT_STORAGE: '2020-10-10'}}
        with mock.patch('azure.cli.core.profiles._shared.AZURE_API_PROFILES', test_profile):
            self.assertEqual(
                get_versioned_sdk_path('latest', ResourceType.MGMT_STORAGE),
                "azure.mgmt.storage.v2020_10_10"
            )


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "policy definition version create",
)
class Create(AAZCommand):
    """Create a policy definition version.

    Create a policy definition version in the given subscription or management group with the given name and other properties.

    :example: Create a policy definition with older version
        az policy definition version create --name readOnlyStorage --rules "{ 'if': { 'field': 'type', 'equals': 'Microsoft.Storage/storageAccounts/write' }, 'then': { 'effect': 'deny' } }" --version 1.1.0
    """

    _aaz_info = {
        "version": "2025-11-01",
        "resources": [
            ["mgmt-plane", "/providers/microsoft.management/managementgroups/{}/providers/microsoft.authorization/policydefinitions/{}/versions/{}", "2025-11-01"],
            ["mgmt-plane", "/subscriptions/{}/providers/microsoft.authorization/policydefinitions/{}/versions/{}", "2025-11-01"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.management_group = AAZStrArg(
            options=["--management-group"],
            help="The management group.",
            fmt=AAZStrArgFormat(
                max_length=90,
                min_length=1,
            ),
        )
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help="The name of the policy definition.",
            required=True,
            fmt=AAZStrArgFormat(
                pattern="^[^<>%&:\\?/]*[^<>%&:\\?/ ]+$",
            ),
        )
        _args_schema.version = AAZStrArg(
            options=["--version"],
            help={"short-summary": "The policy definition version.", "long-summary": "The policy definition version in #.#.# format."},
            required=True,
            fmt=AAZStrArgFormat(
                pattern="^\\d+\\.\\d+\\.\\d+$",
            ),
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.description = AAZStrArg(
            options=["--description"],
            arg_group="Properties",
            help="The policy definition description.",
        )
        _args_schema.display_name = AAZStrArg(
            options=["--display-name"],
            arg_group="Properties",
            help="The display name of the policy definition.",
        )
        _args_schema.external_evaluation_enforcement_settings = AAZObjectArg(
            options=["--external-settings", "--external-evaluation-enforcement-settings"],
            arg_group="Properties",
            help="The details of the source of external evaluation results required by the policy during enforcement evaluation.",
        )
        _args_schema.metadata = AAZAnyTypeArg(
            options=["--metadata"],
            arg_group="Properties",
            help={"short-summary": "The policy definition metadata.", "long-summary": "The policy definition metadata. Metadata is an open-ended object and is typically a collection of key value pairs."},
        )
        _args_schema.mode = AAZStrArg(
            options=["-m", "--mode"],
            arg_group="Properties",
            help={"short-summary": "The policy definition mode.", "long-summary": "The policy definition mode. Valid values for control plane policy definitions: All, Indexed. The mode 'Indexed' indicates the policy should be evaluated only for resource types that support tags and location. Some examples for data plane policy definitions: Microsoft.KeyVault.Data, Microsoft.Network.Data."},
            default="Indexed",
        )
        _args_schema.params = AAZDictArg(
            options=["-p", "--params"],
            arg_group="Properties",
            help={"short-summary": "The policy rule parameter definitions.", "long-summary": "The definitions for parameters used in the policy rule. The keys are the parameter names."},
        )
        _args_schema.rules = AAZAnyTypeArg(
            options=["--rule", "--rules"],
            arg_group="Properties",
            help="The policy rule.",
        )

        external_evaluation_enforcement_settings = cls._args_schema.external_evaluation_enforcement_settings
        external_evaluation_enforcement_settings.endpoint_settings = AAZObjectArg(
            options=["endpoint-settings"],
            help="The settings of an external endpoint providing evaluation results.",
        )
        external_evaluation_enforcement_settings.missing_token_action = AAZStrArg(
            options=["missing-token-action"],
            help="What to do when evaluating an enforcement policy that requires an external evaluation and the token is missing. Possible values are Audit and Deny and language expressions are supported.",
        )
        external_evaluation_enforcement_settings.result_lifespan = AAZStrArg(
            options=["result-lifespan"],
            help="The lifespan of the endpoint invocation result after which it's no longer valid. Value is expected to follow the ISO 8601 duration format and language expressions are supported.",
        )
        external_evaluation_enforcement_settings.role_definition_ids = AAZListArg(
            options=["role-definition-ids"],
            help="An array of the role definition Ids the assignment's MSI will need in order to invoke the endpoint.",
        )

        endpoint_settings = cls._args_schema.external_evaluation_enforcement_settings.endpoint_settings
        endpoint_settings.details = AAZAnyTypeArg(
            options=["details"],
            help="The details of the endpoint.",
        )
        endpoint_settings.kind = AAZStrArg(
            options=["kind"],
            help="The kind of the endpoint.",
        )

        role_definition_ids = cls._args_schema.external_evaluation_enforcement_settings.role_definition_ids
        role_definition_ids.Element = AAZStrArg()

        params = cls._args_schema.params
        params.Element = AAZObjectArg()

        _element = cls._args_schema.params.Element
        _element.allowed_values = AAZListArg(
            options=["allowed-values"],
            help="The allowed values for the parameter.",
        )
        _element.default_value = AAZAnyTypeArg(
            options=["default-value"],
            help="The default value for the parameter if no value is provided.",
        )
        _element.metadata = AAZFreeFormDictArg(
            options=["metadata"],
            help="General metadata for the parameter.",
        )
        _element.schema = AAZAnyTypeArg(
            options=["schema"],
            help="Provides validation of parameter inputs during assignment using a self-defined JSON schema. This property is only supported for object-type parameters and follows the Json.NET Schema 2019-09 implementation. You can learn more about using schemas at https://json-schema.org/ and test draft schemas at https://www.jsonschemavalidator.net/.",
        )
        _element.type = AAZStrArg(
            options=["type"],
            help="The data type of the parameter.",
            enum={"Array": "Array", "Boolean": "Boolean", "DateTime": "DateTime", "Float": "Float", "Integer": "Integer", "Object": "Object", "String": "String"},
        )

        allowed_values = cls._args_schema.params.Element.allowed_values
        allowed_values.Element = AAZAnyTypeArg()
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        condition_0 = has_value(self.ctx.args.management_group) and has_value(self.ctx.args.name) and has_value(self.ctx.args.version)
        condition_1 = has_value(self.ctx.args.name) and has_value(self.ctx.args.version) and has_value(self.ctx.subscription_id)
        if condition_0:
            self.PolicyDefinitionVersionsCreateOrUpdateAtManagementGroup(ctx=self.ctx)()
        if condition_1:
            self.PolicyDefinitionVersionsCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class PolicyDefinitionVersionsCreateOrUpdateAtManagementGroup(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200, 201]:
                return self.on_200_201(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/providers/Microsoft.Management/managementGroups/{managementGroupName}/providers/Microsoft.Authorization/policyDefinitions/{policyDefinitionName}/versions/{policyDefinitionVersion}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "managementGroupName", self.ctx.args.management_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "policyDefinitionName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "policyDefinitionVersion", self.ctx.args.version,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-11-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("description", AAZStrType, ".description")
                properties.set_prop("displayName", AAZStrType, ".display_name")
                properties.set_prop("externalEvaluationEnforcementSettings", AAZObjectType, ".external_evaluation_enforcement_settings")
                properties.set_prop("metadata", AAZAnyType, ".metadata")
                properties.set_prop("mode", AAZStrType, ".mode")
                properties.set_prop("parameters", AAZDictType, ".params")
                properties.set_prop("policyRule", AAZAnyType, ".rules")

            external_evaluation_enforcement_settings = _builder.get(".properties.externalEvaluationEnforcementSettings")
            if external_evaluation_enforcement_settings is not None:
                external_evaluation_enforcement_settings.set_prop("endpointSettings", AAZObjectType, ".endpoint_settings")
                external_evaluation_enforcement_settings.set_prop("missingTokenAction", AAZStrType, ".missing_token_action")
                external_evaluation_enforcement_settings.set_prop("resultLifespan", AAZStrType, ".result_lifespan")
                external_evaluation_enforcement_settings.set_prop("roleDefinitionIds", AAZListType, ".role_definition_ids")

            endpoint_settings = _builder.get(".properties.externalEvaluationEnforcementSettings.endpointSettings")
            if endpoint_settings is not None:
                endpoint_settings.set_prop("details", AAZAnyType, ".details")
                endpoint_settings.set_prop("kind", AAZStrType, ".kind")

            role_definition_ids = _builder.get(".properties.externalEvaluationEnforcementSettings.roleDefinitionIds")
            if role_definition_ids is not None:
                role_definition_ids.set_elements(AAZStrType, ".")

            parameters = _builder.get(".properties.parameters")
            if parameters is not None:
                parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.parameters{}")
            if _elements is not None:
                _elements.set_prop("allowedValues", AAZListType, ".allowed_values")
                _elements.set_prop("defaultValue", AAZAnyType, ".default_value")
                _elements.set_prop("metadata", AAZFreeFormDictType, ".metadata")
                _elements.set_prop("schema", AAZAnyType, ".schema")
                _elements.set_prop("type", AAZStrType, ".type")

            allowed_values = _builder.get(".properties.parameters{}.allowedValues")
            if allowed_values is not None:
                allowed_values.set_elements(AAZAnyType, ".")

            metadata = _builder.get(".properties.parameters{}.metadata")
            if metadata is not None:
                metadata.set_anytype_elements(".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()

            _schema_on_200_201 = cls._schema_on_200_201
            _schema_on_200_201.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200_201.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _schema_on_200_201.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200_201.properties
            properties.description = AAZStrType()
            properties.display_name = AAZStrType(
                serialized_name="displayName",
            )
            properties.external_evaluation_enforcement_settings = AAZObjectType(
                serialized_name="externalEvaluationEnforcementSettings",
            )
            properties.metadata = AAZAnyType()
            properties.mode = AAZStrType()
            properties.parameters = AAZDictType()
            properties.policy_rule = AAZAnyType(
                serialized_name="policyRule",
            )
            properties.policy_type = AAZStrType(
                serialized_name="policyType",
            )
            properties.version = AAZStrType()

            external_evaluation_enforcement_settings = cls._schema_on_200_201.properties.external_evaluation_enforcement_settings
            external_evaluation_enforcement_settings.endpoint_settings = AAZObjectType(
                serialized_name="endpointSettings",
            )
            external_evaluation_enforcement_settings.missing_token_action = AAZStrType(
                serialized_name="missingTokenAction",
            )
            external_evaluation_enforcement_settings.result_lifespan = AAZStrType(
                serialized_name="resultLifespan",
            )
            external_evaluation_enforcement_settings.role_definition_ids = AAZListType(
                serialized_name="roleDefinitionIds",
            )

            endpoint_settings = cls._schema_on_200_201.properties.external_evaluation_enforcement_settings.endpoint_settings
            endpoint_settings.details = AAZAnyType()
            endpoint_settings.kind = AAZStrType()

            role_definition_ids = cls._schema_on_200_201.properties.external_evaluation_enforcement_settings.role_definition_ids
            role_definition_ids.Element = AAZStrType()

            parameters = cls._schema_on_200_201.properties.parameters
            parameters.Element = AAZObjectType()

            _element = cls._schema_on_200_201.properties.parameters.Element
            _element.allowed_values = AAZListType(
                serialized_name="allowedValues",
            )
            _element.default_value = AAZAnyType(
                serialized_name="defaultValue",
            )
            _element.metadata = AAZFreeFormDictType()
            _element.schema = AAZAnyType()
            _element.type = AAZStrType()

            allowed_values = cls._schema_on_200_201.properties.parameters.Element.allowed_values
            allowed_values.Element = AAZAnyType()

            system_data = cls._schema_on_200_201.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            return cls._schema_on_200_201

    class PolicyDefinitionVersionsCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200, 201]:
                return self.on_200_201(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policyDefinitions/{policyDefinitionName}/versions/{policyDefinitionVersion}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "policyDefinitionName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "policyDefinitionVersion", self.ctx.args.version,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-11-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("description", AAZStrType, ".description")
                properties.set_prop("displayName", AAZStrType, ".display_name")
                properties.set_prop("externalEvaluationEnforcementSettings", AAZObjectType, ".external_evaluation_enforcement_settings")
                properties.set_prop("metadata", AAZAnyType, ".metadata")
                properties.set_prop("mode", AAZStrType, ".mode")
                properties.set_prop("parameters", AAZDictType, ".params")
                properties.set_prop("policyRule", AAZAnyType, ".rules")

            external_evaluation_enforcement_settings = _builder.get(".properties.externalEvaluationEnforcementSettings")
            if external_evaluation_enforcement_settings is not None:
                external_evaluation_enforcement_settings.set_prop("endpointSettings", AAZObjectType, ".endpoint_settings")
                external_evaluation_enforcement_settings.set_prop("missingTokenAction", AAZStrType, ".missing_token_action")
                external_evaluation_enforcement_settings.set_prop("resultLifespan", AAZStrType, ".result_lifespan")
                external_evaluation_enforcement_settings.set_prop("roleDefinitionIds", AAZListType, ".role_definition_ids")

            endpoint_settings = _builder.get(".properties.externalEvaluationEnforcementSettings.endpointSettings")
            if endpoint_settings is not None:
                endpoint_settings.set_prop("details", AAZAnyType, ".details")
                endpoint_settings.set_prop("kind", AAZStrType, ".kind")

            role_definition_ids = _builder.get(".properties.externalEvaluationEnforcementSettings.roleDefinitionIds")
            if role_definition_ids is not None:
                role_definition_ids.set_elements(AAZStrType, ".")

            parameters = _builder.get(".properties.parameters")
            if parameters is not None:
                parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.parameters{}")
            if _elements is not None:
                _elements.set_prop("allowedValues", AAZListType, ".allowed_values")
                _elements.set_prop("defaultValue", AAZAnyType, ".default_value")
                _elements.set_prop("metadata", AAZFreeFormDictType, ".metadata")
                _elements.set_prop("schema", AAZAnyType, ".schema")
                _elements.set_prop("type", AAZStrType, ".type")

            allowed_values = _builder.get(".properties.parameters{}.allowedValues")
            if allowed_values is not None:
                allowed_values.set_elements(AAZAnyType, ".")

            metadata = _builder.get(".properties.parameters{}.metadata")
            if metadata is not None:
                metadata.set_anytype_elements(".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()

            _schema_on_200_201 = cls._schema_on_200_201
            _schema_on_200_201.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200_201.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _schema_on_200_201.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200_201.properties
            properties.description = AAZStrType()
            properties.display_name = AAZStrType(
                serialized_name="displayName",
            )
            properties.external_evaluation_enforcement_settings = AAZObjectType(
                serialized_name="externalEvaluationEnforcementSettings",
            )
            properties.metadata = AAZAnyType()
            properties.mode = AAZStrType()
            properties.parameters = AAZDictType()
            properties.policy_rule = AAZAnyType(
                serialized_name="policyRule",
            )
            properties.policy_type = AAZStrType(
                serialized_name="policyType",
            )
            properties.version = AAZStrType()

            external_evaluation_enforcement_settings = cls._schema_on_200_201.properties.external_evaluation_enforcement_settings
            external_evaluation_enforcement_settings.endpoint_settings = AAZObjectType(
                serialized_name="endpointSettings",
            )
            external_evaluation_enforcement_settings.missing_token_action = AAZStrType(
                serialized_name="missingTokenAction",
            )
            external_evaluation_enforcement_settings.result_lifespan = AAZStrType(
                serialized_name="resultLifespan",
            )
            external_evaluation_enforcement_settings.role_definition_ids = AAZListType(
                serialized_name="roleDefinitionIds",
            )

            endpoint_settings = cls._schema_on_200_201.properties.external_evaluation_enforcement_settings.endpoint_settings
            endpoint_settings.details = AAZAnyType()
            endpoint_settings.kind = AAZStrType()

            role_definition_ids = cls._schema_on_200_201.properties.external_evaluation_enforcement_settings.role_definition_ids
            role_definition_ids.Element = AAZStrType()

            parameters = cls._schema_on_200_201.properties.parameters
            parameters.Element = AAZObjectType()

            _element = cls._schema_on_200_201.properties.parameters.Element
            _element.allowed_values = AAZListType(
                serialized_name="allowedValues",
            )
            _element.default_value = AAZAnyType(
                serialized_name="defaultValue",
            )
            _element.metadata = AAZFreeFormDictType()
            _element.schema = AAZAnyType()
            _element.type = AAZStrType()

            allowed_values = cls._schema_on_200_201.properties.parameters.Element.allowed_values
            allowed_values.Element = AAZAnyType()

            system_data = cls._schema_on_200_201.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long

from collections import defaultdict

import argparse
from knack.util import CLIError
from azure.cli.core.azclierror import UnrecognizedArgumentError, InvalidArgumentValueError
from ._validators import read_base_64_file


# pylint: disable=protected-access
class AddBackendAddressCreate(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        action = self.get_action(values, option_string)
        super().__call__(parser, namespace, action, option_string)

    def get_action(self, values, option_string):  # pylint: disable=no-self-use
        try:
            properties = defaultdict(list)
            for (k, v) in (x.split('=', 1) for x in values):
                properties[k].append(v)
            properties = dict(properties)
        except ValueError:
            raise CLIError('usage error: {} [KEY=VALUE ...]'.format(option_string))
        d = {}
        for k in properties:
            kl = k.lower()
            v = properties[k]
            if kl == 'name':
                d['name'] = v[0]
            elif kl == 'ip-address':
                d['ip_address'] = v[0]
            elif kl == 'subnet':
                d['subnet'] = v[0]
            else:
                raise CLIError('key error: key must be one of name, ip-address and subnet.')
        return d


class AddBackendAddressCreateForCrossRegionLB(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        action = self.get_action(values, option_string)
        super().__call__(parser, namespace, action, option_string)

    def get_action(self, values, option_string):  # pylint: disable=no-self-use
        try:
            properties = defaultdict(list)
            for (k, v) in (x.split('=', 1) for x in values):
                properties[k].append(v)
            properties = dict(properties)
        except ValueError:
            raise CLIError('usage error: {} [KEY=VALUE ...]'.format(option_string))
        d = {}
        for k in properties:
            kl = k.lower()
            v = properties[k]
            if kl == 'name':
                d['name'] = v[0]
            elif kl == 'frontend-ip-address':
                d['frontend_ip_address'] = v[0]
            else:
                raise CLIError('key error: key must be one of name and frontend-ip-address.')
        return d


class TrustedClientCertificateCreate(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        action = self.get_action(values, option_string)
        super().__call__(parser, namespace, action, option_string)

    def get_action(self, values, option_string):  # pylint: disable=no-self-use
        try:
            properties = defaultdict(list)
            for (k, v) in (x.split('=', 1) for x in values):
                properties[k].append(v)
            properties = dict(properties)
        except ValueError:
            raise UnrecognizedArgumentError('usage error: {} [KEY=VALUE ...]'.format(option_string))
        d = {}
        for k in properties:
            kl = k.lower()
            v = properties[k]
            if kl == 'name':
                d['name'] = v[0]
            elif kl == 'data':
                d['data'] = read_base_64_file(v[0])
            else:
                raise UnrecognizedArgumentError('key error: key must be one of name and data.')
        return d


def _split(param):
    return param.split(',')


class SslProfilesCreate(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        action = self.get_action(values, option_string)
        super().__call__(parser, namespace, action, option_string)

    def get_action(self, values, option_string):  # pylint: disable=no-self-use
        try:
            properties = defaultdict(list)
            for (k, v) in (x.split('=', 1) for x in values):
                properties[k].append(v)
            properties = dict(properties)
        except ValueError:
            raise UnrecognizedArgumentError('usage error: {} [KEY=VALUE ...]'.format(option_string))
        d = {}
        for k in properties:
            kl = k.lower()
            v = properties[k]
            if kl == 'name':
                d['name'] = v[0]
            elif kl == 'policy-type':
                d['policy_type'] = v[0]
            elif kl == 'min-protocol-version':
                d['min_protocol_version'] = v[0]
            elif kl == 'cipher-suites':
                d['cipher_suites'] = _split(v[0])
            elif kl == 'disabled-ssl-protocols':
                d['disabled_ssl_protocols'] = _split(v[0])
            elif kl == 'client-auth-configuration':
                d['client_auth_configuration'] = bool(v[0])
            elif kl == 'trusted-client-certificates':
                d['trusted_client_certificates'] = _split(v[0])
            else:
                raise UnrecognizedArgumentError('key error: key must be one of policy-type, min-protocol-version, '
                                                'cipher-suites, client-auth-configuration, trusted-client-certificates,'
                                                'disabled-ssl-protocols.')
        return d


class AddMappingRequest(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        action = self.get_action(values, option_string)
        namespace.request = action

    def get_action(self, values, option_string):  # pylint: disable=no-self-use
        try:
            properties = defaultdict(list)
            for (k, v) in (x.split('=', 1) for x in values):
                properties[k].append(v)
            properties = dict(properties)
        except ValueError:
            raise UnrecognizedArgumentError('Usage error: {} [KEY=VALUE ...]'.format(option_string))
        d = {}
        for k in properties:
            kl = k.lower()
            v = properties[k]
            if kl == 'ip':
                d['ip_address'] = v[0]
            elif kl == 'nic':
                d['ip_configuration'] = v[0]
            else:
                raise UnrecognizedArgumentError('key error: key must be one of {ip, nic}.')
        return d


class WAFRulesCreate(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        action = self.get_action(values, option_string)
        super().__call__(parser, namespace, action, option_string)

    def get_action(self, values, option_string):  # pylint: disable=no-self-use
        try:
            properties = defaultdict(list)
            for (k, v) in (x.split('=', 1) for x in values):
                properties[k].append(v)
            properties = dict(properties)
        except ValueError:
            raise UnrecognizedArgumentError('usage error: {} [KEY=VALUE ...]'.format(option_string))
        d = {}
        for k in properties:
            kl = k.lower()
            v = properties[k]
            if kl == 'rule-id':
                d['rule_id'] = v[0]
            elif kl == 'action':
                d['action'] = v[0]
            elif kl == 'state':
                d['state'] = v[0]
            elif kl == 'sensitivity':
                if v[0].lower() not in ['low', 'medium', 'high']:
                    raise InvalidArgumentValueError("sensitivity must be one of 'low', 'medium', or 'high'.")
                d['sensitivity'] = v[0]
            else:
                raise UnrecognizedArgumentError('key error: key must be one of rule-id, action, state, and sensitivity.')
        return d



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""Generate the Homebrew cask from a template.

Usage:
    python3 scripts/release/macos/cask_generate.py \
        --version "2.x.y" \
        --arm64-sha "<arm64 sha256>" \
        --x86-64-sha "<x86_64 sha256>" \
        --github-repo "Azure/azure-cli" \
        --template "scripts/release/macos/templates/azure-cli.rb.in" \
        --output "azure-cli.rb"

Environment variable fallbacks (used when a CLI flag is omitted):
    VERSION, ARM64_SHA, X86_64_SHA, GITHUB_REPO, TEMPLATE, OUTPUT
"""

import argparse
import os
from pathlib import Path
from typing import Dict

DEFAULT_TEMPLATE = "scripts/release/macos/templates/azure-cli.rb.in"
DEFAULT_OUTPUT = "azure-cli.rb"


def _env_or_arg(value: str, env_key: str) -> str:
    return value if value is not None else os.environ.get(env_key)


def _require(value: str, name: str) -> str:
    if not value:
        raise SystemExit(f"Missing required value: {name}")
    return value


def _render_template(template_path: Path, replacements: Dict[str, str]) -> str:
    content = template_path.read_text()
    for key, value in replacements.items():
        content = content.replace(key, value)

    if "{{" in content:
        raise SystemExit("Template placeholders not fully replaced")
    if not content.endswith("\n"):
        content += "\n"
    return content


def generate_cask(args: argparse.Namespace) -> None:
    version = _require(_env_or_arg(args.version, "VERSION"), "version")
    arm64_sha = _require(_env_or_arg(args.arm64_sha, "ARM64_SHA"), "arm64_sha")
    x86_64_sha = _require(_env_or_arg(args.x86_64_sha, "X86_64_SHA"), "x86_64_sha")
    github_repo = _require(_env_or_arg(args.github_repo, "GITHUB_REPO"), "github_repo")
    python_version = _require(_env_or_arg(args.python_version, "PYTHON_VERSION"), "python_version")

    template_path = Path(_env_or_arg(args.template, "TEMPLATE") or DEFAULT_TEMPLATE)
    output_path = Path(_env_or_arg(args.output, "OUTPUT") or DEFAULT_OUTPUT)

    replacements = {
        "{{ version }}": version,
        "{{ arm64_sha }}": arm64_sha,
        "{{ x86_64_sha }}": x86_64_sha,
        "{{ github_repo }}": github_repo,
        "{{ python_version }}": python_version,
    }

    content = _render_template(template_path, replacements)
    output_path.write_text(content)


def main() -> None:
    parser = argparse.ArgumentParser(prog="cask_generate.py")
    parser.add_argument("--version", dest="version", help="Azure CLI version")
    parser.add_argument("--arm64-sha", dest="arm64_sha", help="ARM64 tarball SHA256")
    parser.add_argument("--x86-64-sha", dest="x86_64_sha", help="x86_64 tarball SHA256")
    parser.add_argument("--github-repo", dest="github_repo", help="GitHub repo, e.g. Azure/azure-cli")
    parser.add_argument("--python-version", dest="python_version", help="Python major.minor version, e.g. 3.14")
    parser.add_argument("--template", dest="template", help="Template path (.rb.in)")
    parser.add_argument("--output", dest="output", help="Output cask path (.rb)")
    parser.set_defaults(func=generate_cask)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=protected-access

from knack.log import get_logger

logger = get_logger(__name__)


# region ActivityLog
def list_activity_log(cmd, correlation_id=None, resource_group=None, resource_id=None,
                      resource_provider=None, start_time=None, end_time=None, caller=None, status=None, max_events=50,
                      select=None, offset='6h'):
    odata_filters = _build_activity_log_odata_filter(correlation_id, resource_group, resource_id, resource_provider,
                                                     start_time, end_time, caller, status, offset)

    select_filters = _activity_log_select_filter_builder(select)
    logger.info('OData Filter: %s', odata_filters)
    logger.info('Select Filter: %s', select_filters)

    from .aaz.latest.monitor.activity_log._list import List
    # activity_log = client.list(filter=odata_filters, select=select_filters)
    activity_log = List(cli_ctx=cmd.cli_ctx)(command_args={
        "filter": odata_filters,
        "select": select_filters,
    })

    return _limit_results(activity_log, max_events)


def _build_activity_log_odata_filter(correlation_id=None, resource_group=None, resource_id=None, resource_provider=None,
                                     start_time=None, end_time=None, caller=None, status=None, offset=None):
    from datetime import datetime
    import dateutil.parser

    if not start_time and not end_time:
        # if neither value provided, end_time is now
        end_time = datetime.utcnow().isoformat()
    if not start_time:
        # if no start_time, apply offset backwards from end_time
        start_time = (dateutil.parser.parse(end_time) - offset).isoformat()
    elif not end_time:
        # if no end_time, apply offset fowards from start_time
        end_time = (dateutil.parser.parse(start_time) + offset).isoformat()

    odata_filters = 'eventTimestamp ge {} and eventTimestamp le {}'.format(start_time, end_time)

    if correlation_id:
        odata_filters = _build_odata_filter(odata_filters, 'correlation_id', correlation_id, 'correlationId')
    elif resource_group:
        odata_filters = _build_odata_filter(odata_filters, 'resource_group', resource_group, 'resourceGroupName')
    elif resource_id:
        odata_filters = _build_odata_filter(odata_filters, 'resource_id', resource_id, 'resourceId')
    elif resource_provider:
        odata_filters = _build_odata_filter(odata_filters, 'resource_provider', resource_provider, 'resourceProvider')
    if caller:
        odata_filters = _build_odata_filter(odata_filters, 'caller', caller, 'caller')
    if status:
        odata_filters = _build_odata_filter(odata_filters, 'status', status, 'status')

    return odata_filters


def _activity_log_select_filter_builder(events=None):
    """Build up select filter string from events"""
    if events:
        return ' , '.join(events)
    return None


def _build_odata_filter(default_filter, field_name, field_value, field_label):
    if not field_value:
        from knack.util import CLIError
        raise CLIError('Value for {} can not be empty.'.format(field_name))

    return _add_condition(default_filter, field_label, field_value)


def _add_condition(default_filter, field_label, field_value):
    if not field_value:
        return default_filter

    return "{} and {} eq '{}'".format(default_filter, field_label, field_value)


def _single(collection):
    return len([x for x in collection if x]) == 1


def _limit_results(paged, limit):
    results = []
    for index, item in enumerate(paged):
        if index < limit:
            results.append(item)
        else:
            break
    return list(results)
# endregion


# region Metrics
# pylint:disable=unused-argument
def list_metrics(cmd, resource,
                 start_time=None, end_time=None, offset='1h', interval='1m',
                 metadata=None, dimension=None, aggregation=None, metrics=None,
                 filters=None, metric_namespace=None, orderby=None, top=10):

    from datetime import datetime
    import dateutil.parser
    from urllib.parse import quote_plus

    if not start_time and not end_time:
        # if neither value provided, end_time is now
        end_time = datetime.utcnow().isoformat()
    if not start_time:
        # if no start_time, apply offset backwards from end_time
        start_time = (dateutil.parser.parse(end_time) - offset).isoformat()
    elif not end_time:
        # if no end_time, apply offset fowards from start_time
        end_time = (dateutil.parser.parse(start_time) + offset).isoformat()

    timespan = '{}/{}'.format(start_time, end_time)

    from .aaz.latest.monitor.metrics._list import List
    return List(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_uri": resource,
        "timespan": quote_plus(timespan),
        "interval": interval,
        "metricnames": ','.join(metrics) if metrics else None,
        "aggregation": ','.join(aggregation) if aggregation else None,
        "top": top,
        "orderby": orderby,
        "filter": filters,
        "result_typ": "Metadata" if metadata else None,
        "metricnamespac": metric_namespace
    })


def list_definations(cmd, resource_uri, metricnamespace=None):
    from .aaz.latest.monitor.metrics._list_definitions import ListDefinitions
    return ListDefinitions(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_uri": resource_uri,
        "metricnamespace": metricnamespace
    })


def list_namespaces(cmd, resource_uri, start_time=None):
    from .aaz.latest.monitor.metrics._list_namespaces import ListNamespaces
    return ListNamespaces(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_uri": resource_uri,
        "start_time": start_time
    })
# endregion



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=unused-argument, line-too-long
from azure.cli.core.util import CLIError
from azure.mgmt import postgresqlflexibleservers as postgresql_flexibleservers
from .._client_factory import cf_postgres_flexible_config
from ..utils._flexible_server_util import resolve_poller
from ..utils.validators import validate_citus_cluster, validate_resource_group


def flexible_server_identity_update(cmd, client, resource_group_name, server_name, system_assigned):
    validate_resource_group(resource_group_name)
    validate_citus_cluster(cmd, resource_group_name, server_name)

    server = client.get(resource_group_name, server_name)
    identity_type = server.identity.type if (server and server.identity and server.identity.type) else 'None'

    if system_assigned.lower() == 'enabled':
        # user wants to enable system-assigned identity
        if identity_type == 'None':
            # if user-assigned identity is not enabled, then enable system-assigned identity
            identity_type = 'SystemAssigned'
        elif identity_type == 'UserAssigned':
            # if user-assigned identity is enabled, then enable both system-assigned and user-assigned identity
            identity_type = 'SystemAssigned,UserAssigned'
    else:
        # check if fabric is enabled
        config_client = cf_postgres_flexible_config(cmd.cli_ctx, '_')
        fabric_mirror_status = config_client.get(resource_group_name, server_name, 'azure.fabric_mirror_enabled')
        if (fabric_mirror_status and fabric_mirror_status.value.lower() == 'on'):
            raise CLIError("On servers for which Fabric mirroring is enabled, system assigned managed identity cannot be disabled.")
        if server.data_encryption.type == 'AzureKeyVault':
            # if data encryption is enabled, then system-assigned identity cannot be disabled
            raise CLIError("On servers for which data encryption is based on customer managed key, system assigned managed identity cannot be disabled.")
        if identity_type == 'SystemAssigned,UserAssigned':
            # if both system-assigned and user-assigned identity is enabled, then disable system-assigned identity
            identity_type = 'UserAssigned'
        elif identity_type == 'SystemAssigned':
            # if only system-assigned identity is enabled, then disable system-assigned identity
            identity_type = 'None'

    if identity_type == 'UserAssigned' or identity_type == 'SystemAssigned,UserAssigned':
        identities_map = {}
        for identity in server.identity.user_assigned_identities:
            identities_map[identity] = {}
        parameters = {
            'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
                user_assigned_identities=identities_map,
                type=identity_type)}
    else:
        parameters = {
            'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
                type=identity_type)}

    result = resolve_poller(
        client.begin_update(
            resource_group_name=resource_group_name,
            server_name=server_name,
            parameters=parameters),
        cmd.cli_ctx, 'Updating user assigned identity type for server {}'.format(server_name)
    )

    return result.identity


def flexible_server_identity_assign(cmd, client, resource_group_name, server_name, identities):
    validate_resource_group(resource_group_name)
    validate_citus_cluster(cmd, resource_group_name, server_name)

    server = client.get(resource_group_name, server_name)
    identity_type = server.identity.type if (server and server.identity and server.identity.type) else 'None'

    if identity_type == 'SystemAssigned':
        # if system-assigned identity is enabled, then enable both system
        identity_type = 'SystemAssigned,UserAssigned'
    elif identity_type == 'None':
        # if system-assigned identity is not enabled, then enable user-assigned identity
        identity_type = 'UserAssigned'

    identities_map = {}
    for identity in identities:
        identities_map[identity] = {}

    parameters = {
        'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
            user_assigned_identities=identities_map,
            type=identity_type)}

    result = resolve_poller(
        client.begin_update(
            resource_group_name=resource_group_name,
            server_name=server_name,
            parameters=parameters),
        cmd.cli_ctx, 'Adding identities to server {}'.format(server_name)
    )

    return result.identity


def flexible_server_identity_remove(cmd, client, resource_group_name, server_name, identities):
    validate_resource_group(resource_group_name)
    validate_citus_cluster(cmd, resource_group_name, server_name)

    instance = client.get(resource_group_name, server_name)

    if instance.data_encryption:
        primary_id = instance.data_encryption.primary_user_assigned_identity_id

        if primary_id and primary_id.lower() in [identity.lower() for identity in identities]:
            raise CLIError("Cannot remove identity {} because it's used for data encryption.".format(primary_id))

        geo_backup_id = instance.data_encryption.geo_backup_user_assigned_identity_id

        if geo_backup_id and geo_backup_id.lower() in [identity.lower() for identity in identities]:
            raise CLIError("Cannot remove identity {} because it's used for geo backup data encryption.".format(geo_backup_id))

    identities_map = {}
    for identity in identities:
        identities_map[identity] = None

    system_assigned_identity = instance.identity and instance.identity.principal_id is not None

    # if there are no user-assigned identities or all user-assigned identities are already removed
    if not (instance.identity and instance.identity.user_assigned_identities) or \
       all(key.lower() in [identity.lower() for identity in identities] for key in instance.identity.user_assigned_identities.keys()):
        if system_assigned_identity:
            # if there is system assigned identity, then set identity type to SystemAssigned
            parameters = {
                'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
                    type="SystemAssigned")}
        else:
            # no system assigned identity, set identity type to None
            parameters = {
                'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
                    type="None")}
    # if there are user-assigned identities and system assigned identity, then set identity type to SystemAssigned,UserAssigned
    elif system_assigned_identity:
        parameters = {
            'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
                user_assigned_identities=identities_map,
                type="SystemAssigned,UserAssigned")}
    # there is no system assigned identity, but there are user-assigned identities, then set identity type to UserAssigned
    else:
        parameters = {
            'identity': postgresql_flexibleservers.models.UserAssignedIdentity(
                user_assigned_identities=identities_map,
                type="UserAssigned")}

    result = resolve_poller(
        client.begin_update(
            resource_group_name=resource_group_name,
            server_name=server_name,
            parameters=parameters),
        cmd.cli_ctx, 'Removing identities from server {}'.format(server_name)
    )

    return result.identity or postgresql_flexibleservers.models.UserAssignedIdentity(type="SystemAssigned")


def flexible_server_identity_list(cmd, client, resource_group_name, server_name):
    validate_resource_group(resource_group_name)
    validate_citus_cluster(cmd, resource_group_name, server_name)

    server = client.get(resource_group_name, server_name)
    return server.identity or postgresql_flexibleservers.models.UserAssignedIdentity(type="SystemAssigned")


def flexible_server_identity_show(cmd, client, resource_group_name, server_name, identity):
    validate_resource_group(resource_group_name)
    validate_citus_cluster(cmd, resource_group_name, server_name)

    server = client.get(resource_group_name, server_name)

    for key, value in server.identity.user_assigned_identities.items():
        if key.lower() == identity.lower():
            return value

    raise CLIError("Identity '{}' does not exist in server {}.".format(identity, server_name))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "monitor log-analytics workspace failover",
)
class Failover(AAZCommand):
    """Activates failover for the specified workspace.The specified replication location must match the location of the enabled replication for this workspace.The failover operation is asynchronous and can take up to 30 minutes to complete.The status of the operation can be checked using the operationId returned in the response.

    :example: Activates failover for the specified workspace
        az monitor log-analytics workspace failover --resource-group oiautorest6685 --location eastus --workspace-name oiautorest6685
    """

    _aaz_info = {
        "version": "2025-02-01",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.operationalinsights/locations/{}/workspaces/{}/failover", "2025-02-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, None)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.location = AAZResourceLocationArg(
            required=True,
            id_part="name",
            fmt=AAZResourceLocationArgFormat(
                resource_group_arg="resource_group",
            ),
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.workspace_name = AAZStrArg(
            options=["--name", "--workspace-name"],
            help="The name of the workspace.",
            required=True,
            id_part="child_name_1",
            fmt=AAZStrArgFormat(
                pattern="^[A-Za-z0-9][A-Za-z0-9-]+[A-Za-z0-9]$",
                max_length=63,
                min_length=4,
            ),
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.WorkspacesFailover(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    class WorkspacesFailover(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.OperationalInsights/locations/{location}/workspaces/{workspaceName}/failover",
                **self.url_parameters
            )

        @property
        def method(self):
            return "POST"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "location", self.ctx.args.location,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
                **self.serialize_url_param(
                    "workspaceName", self.ctx.args.workspace_name,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-02-01",
                    required=True,
                ),
            }
            return parameters

        def on_200(self, session):
            pass


class _FailoverHelper:
    """Helper class for Failover"""


__all__ = ["Failover"]
